from enum import Enum


class Routes(str, Enum):
    healthcheck = "healthcheck"
    username_login = "login"
    api_key_login = "login/api_key"
    get_token = "get-token"
    current_user = "current_user"
    projects = "projects"
    project = "projects/{project_id}"
    project_groups = "projects/{project_id}/groups"
    project_users = "projects/{project_id}/users"
    groups = "groups"
    group_members = "groups/{group_id}/members"
    invite_users = "invite_users"
    users = "users"
    create_api_key = "users/api_keys"
    get_api_keys = "users/{user_id}/api_keys"
    delete_api_key = "users/api_keys/{api_key_id}"
    update_user = "users/{user_id}"
    datasets = "datasets"
    dataset_content = "datasets/{dataset_id}/content"



from pydantic import UUID4

from galileo_core.constants.request_method import RequestMethod
from galileo_core.constants.routes import Routes
from galileo_core.helpers.logger import logger
from galileo_core.schemas.base_config import BaseConfig
from galileo_core.schemas.core.collaboration_role import CollaboratorRole
from galileo_core.schemas.core.group_project import GroupProjectCollaboratorRequest, GroupProjectCollaboratorResponse


def share_project_with_group(
    project_id: UUID4, group_id: UUID4, role: CollaboratorRole = CollaboratorRole.viewer
) -> GroupProjectCollaboratorResponse:
    config = BaseConfig.get()
    logger.debug(f"Sharing project {project_id} with group {group_id} with role {role}...")
    response_dict = config.api_client.request(
        RequestMethod.POST,
        Routes.project_groups.format(project_id=project_id),
        json=[GroupProjectCollaboratorRequest(group_id=group_id, role=role).model_dump(mode="json")],
    )
    group_collaborators = [GroupProjectCollaboratorResponse.model_validate(group) for group in response_dict]
    logger.debug(f"Shared project {project_id} with group {group_id} with role {role}.")
    return group_collaborators[0]



from asyncio import new_event_loop, run_coroutine_threadsafe
from threading import Thread
from typing import Any, Awaitable, Coroutine, TypeVar

T = TypeVar("T")


class AsyncExecutor:
    # Copied from: https://gist.github.com/gsakkis/18bc444607a590fe3f084a77aa54b4c2
    def __enter__(self) -> "AsyncExecutor":
        self._loop = new_event_loop()
        self._looper = Thread(target=self._loop.run_forever, daemon=True)
        self._looper.start()
        return self

    def __call__(self, coro: Awaitable[T]) -> T:
        return run_coroutine_threadsafe(coro, self._loop).result()

    def __exit__(self, *exc_info: Any) -> None:
        self._loop.call_soon_threadsafe(self._loop.stop)
        self._looper.join()
        self._loop.close()


def async_run(coroutine: Coroutine) -> Any:
    """
    Run an async coroutine synchronously.

    This function is useful for running async code in a synchronous context. This will
    ensure that the async code is run in a separate thread and the result is returned
    to the caller.

    Parameters
    ----------
    coroutine : Coroutine
        The coroutine to run.

    Returns
    -------
    Any
        The result of the coroutine.
    """
    with AsyncExecutor() as async_executor:
        return async_executor(coroutine)



from functools import partial
from typing import List, Optional

from pydantic import UUID4

from galileo_core.constants.request_method import RequestMethod
from galileo_core.constants.routes import Routes
from galileo_core.helpers.logger import logger
from galileo_core.helpers.pagination import paginated_request
from galileo_core.schemas.base_config import BaseConfig
from galileo_core.schemas.core.group import (
    AddGroupMemberRequest,
    AddGroupMemberResponse,
    CreateGroupRequest,
    CreateGroupResponse,
)
from galileo_core.schemas.core.group_role import GroupRole
from galileo_core.schemas.core.group_visibility import GroupVisibility


def create_group(
    name: str, description: Optional[str] = None, visibility: GroupVisibility = GroupVisibility.public
) -> CreateGroupResponse:
    """
    Create a group.

    Parameters
    ----------
    name : str
        Name of the group.
    description : Optional[str], optional
        Description for the group, by default None
    visibility : GroupVisibility, optional
        Visiblity of the group, by default GroupVisibility.public

    Returns
    -------
    CreateGroupResponse
        Response object for the created group.
    """
    config = BaseConfig.get()
    request = CreateGroupRequest(name=name, description=description, visibility=visibility)
    logger.debug(f"Creating group {request.name} with visibility {visibility}...")
    response_dict = config.api_client.request(RequestMethod.POST, Routes.groups, json=request.model_dump(mode="json"))
    group_response = CreateGroupResponse.model_validate(response_dict)
    logger.debug(f"Created group with name {group_response.name}, ID {group_response.id}.")
    return group_response


def list_groups() -> List[CreateGroupResponse]:
    """
    List all groups for the user.

    Returns
    -------
    List[CreateGroupResponse]
        List of all groups.
    """
    config = BaseConfig.get()
    logger.debug("Listing groups...")
    all_groups = paginated_request(partial(config.api_client.request, RequestMethod.GET, Routes.groups), "groups")
    groups = [CreateGroupResponse.model_validate(group) for group in all_groups]
    logger.debug(f"Listed all groups, found {len(groups)} groups.")
    return groups


def add_users_to_group(
    group_id: UUID4, user_ids: List[UUID4], role: GroupRole = GroupRole.member
) -> List[AddGroupMemberResponse]:
    """
    Add users to a an existing group with the specified role.

    Parameters
    ----------
    group_id : UUID4
        Group ID.
    user_ids : List[UUID4]
        List of user IDs to add to the group.
    role : GroupRole
        Role of the user in the group, by default GroupRole.member.

    Returns
    -------
    List[AddGroupMemberResponse]
        List of responses for each user added to the group.
    """
    config = BaseConfig.get()
    request = [AddGroupMemberRequest(user_id=user_id, role=role) for user_id in user_ids]
    logger.debug(f"Adding {len(request)} users to group {group_id} with role {role}...")
    response = config.api_client.request(
        RequestMethod.POST,
        Routes.group_members.format(group_id=group_id),
        json=[r.model_dump(mode="json") for r in request],
    )
    parsed_response = [AddGroupMemberResponse.model_validate(r) for r in response]
    logger.debug(f"Added {len(response)} users to group {group_id} with role {role}.")
    return parsed_response



from typing import Any, Dict
from urllib.parse import urljoin

from httpx import AsyncClient, HTTPError, Response, Timeout
from pydantic import BaseModel, HttpUrl, SecretStr

from galileo_core.constants.http_headers import HttpHeaders
from galileo_core.constants.request_method import RequestMethod
from galileo_core.exceptions.http import GalileoHTTPException
from galileo_core.helpers.execution import async_run
from galileo_core.helpers.logger import logger


class ApiClient(BaseModel):
    host: HttpUrl
    jwt_token: SecretStr
    skip_ssl_validation: bool = False

    @property
    def auth_header(self) -> Dict[str, str]:
        return {"Authorization": f"Bearer {self.jwt_token.get_secret_value()}"}

    @staticmethod
    def validate_response(response: Response) -> None:
        for header, value in response.headers.items():
            # Log all Galileo headers. These are headers that start with `Galileo-Request-`.
            # These can be in any case, so we lower-case them for comparison.
            if header.lower().startswith("galileo-request-"):
                logger.debug(f"{header.title()}: {value}.")
        try:
            response.raise_for_status()
        except HTTPError:
            raise GalileoHTTPException(
                f"Galileo API returned HTTP status code {response.status_code}. Error was: {response.text}",
                response.status_code,
                response.text,
            )

    @staticmethod
    async def make_request(
        request_method: RequestMethod,
        base_url: str,
        endpoint: str,
        skip_ssl_validation: bool = False,
        read_timeout: float = 60.0,
        **kwargs: Any,
    ) -> Any:
        url = urljoin(base_url, endpoint)
        logger.debug(f"Making request to {url}.")
        async with AsyncClient(
            base_url=base_url, timeout=Timeout(read_timeout, connect=5.0), verify=not skip_ssl_validation
        ) as client:
            response = await client.request(method=request_method.value, url=url, timeout=read_timeout, **kwargs)
            ApiClient.validate_response(response)
            logger.debug(f"Response was received from {url}.")
            return response.json()

    def request(
        self,
        method: RequestMethod,
        path: str,
        content_headers: Dict[str, str] = HttpHeaders.json(),
        **kwargs: Any,
    ) -> Any:
        return async_run(self.arequest(method=method, path=path, content_headers=content_headers, **kwargs))

    async def arequest(
        self,
        method: RequestMethod,
        path: str,
        content_headers: Dict[str, str] = HttpHeaders.json(),
        **kwargs: Any,
    ) -> Any:
        return await ApiClient.make_request(
            request_method=method,
            base_url=self.host.unicode_string(),
            endpoint=path,
            headers={**content_headers, **self.auth_header},
            skip_ssl_validation=self.skip_ssl_validation,
            **kwargs,
        )



from typing import Optional, Sequence

from pydantic import BaseModel, ConfigDict, Field, field_validator

from galileo_core.schemas.protect.action import Action, BaseAction, PassthroughAction
from galileo_core.schemas.protect.rule import Rule


class Ruleset(BaseModel):
    rules: Sequence[Rule] = Field(
        description="List of rules to evaluate. Atleast 1 rule is required.",
        default_factory=list,
        min_length=1,
    )
    action: Action = Field(
        description="Action to take if all the rules are met.",
        default_factory=PassthroughAction,
    )
    description: Optional[str] = Field(description="Description of the ruleset.", default=None)

    @field_validator("action", mode="before")
    def validate_action(cls, value: Optional[BaseAction]) -> BaseAction:
        if not value:
            value = PassthroughAction()
        return value


class RulesetsMixin(BaseModel):
    rulesets: Sequence[Ruleset] = Field(
        default_factory=list,
        description="Rulesets to be applied to the payload.",
        validation_alias="prioritized_rulesets",
    )

    model_config = ConfigDict(populate_by_name=True)



from enum import Enum


class ProcessingHeaders(str, Enum):
    received_at = "Galileo-Request-Received-At-Nanoseconds"
    execution_time = "Galileo-Request-Execution-Time-Seconds"
    response_at = "Galileo-Request-Response-At-Nanoseconds"



from enum import Enum
from typing import Optional

from pydantic import BaseModel, field_serializer, field_validator

from galileo_core.schemas.shared.metric import MetricValueType


class MetricComputationStatus(str, Enum):
    success = "success"
    timeout = "timeout"
    failed = "failed"
    error = "error"


class MetricComputation(BaseModel):
    value: MetricValueType = None
    execution_time: Optional[float] = None
    status: Optional[MetricComputationStatus] = None
    error_message: Optional[str] = None

    @field_serializer("status", when_used="json-unless-none")
    def upper_case_status(self, value: MetricComputationStatus) -> str:
        return value.value.upper()

    @field_validator("status", mode="before")
    def lower_case_status(cls, value: Optional[str]) -> Optional[str]:
        if value is None:
            return None
        return value.lower()



from datetime import datetime

SCORER_NAME_REGEX = r"^[\w -]+$"


def ts_name(prefix: str) -> str:
    ts = datetime.now()
    ts_string = ts.strftime("%b_%d_%H_%M_%S")
    return f"{prefix}-{ts_string}"



from time import time_ns
from uuid import uuid4

from pydantic import (
    UUID4,
    BaseModel,
    ConfigDict,
    Field,
    ValidationInfo,
    field_validator,
)

from galileo_core.schemas.protect.execution_status import ExecutionStatus, ExecutionStatusMixIn


class BaseTraceMetadata(BaseModel):
    id: UUID4 = Field(default_factory=uuid4, description="Unique identifier for the request.")
    received_at: int = Field(
        default_factory=time_ns,
        description="Time the request was received by the server in nanoseconds.",
    )


class TraceMetadata(BaseTraceMetadata):
    response_at: int = Field(
        default_factory=time_ns,
        description="Time the response was sent by the server in nanoseconds.",
    )
    # Set a default value for execution_time to -1 to indicate that it is not set.
    execution_time: float = Field(
        default=-1,
        description="Execution time for the request (in seconds).",
        validate_default=True,
    )

    @field_validator("execution_time", mode="before")
    def set_execution_time(cls, value: int, info: ValidationInfo) -> int:
        # Our timestamps are in nanoseconds, so we convert to seconds to get the
        # execution time.
        if value <= 0:
            value = (info.data.get("response_at", 0) - info.data.get("received_at", 0)) * 1e-9
        return value


class Response(ExecutionStatusMixIn):
    text: str = Field(description="Text from the request after processing the rules.")
    trace_metadata: TraceMetadata
    status: ExecutionStatus = Field(
        default=ExecutionStatus.not_triggered, description="Status of the request after processing the rules."
    )

    model_config = ConfigDict(extra="allow")



from enum import Enum
from typing import Union

from pydantic import (
    BaseModel,
    Field,
    field_serializer,
    field_validator,
)


class ExecutionStatus(str, Enum):
    """Status of the execution."""

    triggered = "triggered"
    failed = "failed"
    error = "error"
    timeout = "timeout"
    paused = "paused"
    not_triggered = "not_triggered"


class ExecutionStatusMixIn(BaseModel):
    status: ExecutionStatus = Field(default=ExecutionStatus.not_triggered, description="Status of the execution.")

    @field_validator("status", mode="before")
    def case_agnostic_status(cls, status: Union[str, ExecutionStatus]) -> ExecutionStatus:
        # Ensure that the status is case agnostic.
        return ExecutionStatus(status.lower())

    @field_serializer("status", when_used="always")
    def upper_case_status(self, execution_status: ExecutionStatus) -> str:
        return execution_status.value.upper()



from http.client import HTTPException

from galileo_core.helpers.logger import logger


class GalileoHTTPException(HTTPException):
    """Galileo HTTP exception to wrap all http exceptions."""

    def __init__(self, message: str, status_code: int, response_text: str) -> None:
        self.message = message
        self.status_code = status_code
        self.response_text = response_text
        logger.error(f"Galileo API returned HTTP status code {status_code}. Error was: {response_text}.")



from typing import List

from pydantic import UUID4, BaseModel, Field

from galileo_core.schemas.core.user_role import UserRole


class InviteUsersRequest(BaseModel):
    emails: List[str] = Field(default_factory=list)
    role: UserRole = UserRole.user
    group_ids: List[UUID4] = []


class User(BaseModel):
    id: UUID4
    email: str
    role: UserRole = UserRole.user


class UpdateUserRoleRequest(BaseModel):
    role: UserRole



from logging import Formatter, Logger, StreamHandler, getLogger
from sys import stderr


def get_logger() -> Logger:
    logger = getLogger("galileo_core")
    handler = StreamHandler(stderr)
    handler.setFormatter(Formatter("%(asctime)s.%(msecs)03d | %(levelname)s | %(message)s"))
    if logger.handlers:
        logger.handlers.clear()
    logger.addHandler(handler)
    return logger


logger = get_logger()



from functools import partial
from typing import Dict, List
from uuid import UUID

from galileo_core.constants.request_method import RequestMethod
from galileo_core.constants.routes import Routes
from galileo_core.helpers.logger import logger
from galileo_core.helpers.pagination import paginated_request
from galileo_core.schemas.base_config import BaseConfig
from galileo_core.schemas.core.dataset import Dataset, UploadDatasetRequest
from galileo_core.utils.dataset import DatasetType


def create_dataset(
    dataset: DatasetType,
) -> Dataset:
    """
    Create a dataset.

    Parameters
    ----------
    dataset : DatasetType
        A dataset file path, dictionary, or list of dictionaries.

    Returns
    -------
    Dataset
        Response object for the created dataset.
    """
    config = BaseConfig.get()
    request = UploadDatasetRequest(file_path=dataset)
    logger.debug(f"Creating dataset from {request.file_path}...")
    content_headers = {
        "Accept": "application/json",
    }

    response_dict = config.api_client.request(
        RequestMethod.POST,
        Routes.datasets,
        content_headers=content_headers,
        params=request.params,
        files=request.files,
    )
    dataset_response = Dataset.model_validate(response_dict)
    logger.debug(f"Created dataset with name {dataset_response.name}, ID {dataset_response.id}.")
    return dataset_response


def list_datasets() -> List[Dataset]:
    """
    List all datasets for the user.

    Returns
    -------
    List[Dataset]
        List of all datasets.
    """
    config = BaseConfig.get()
    logger.debug("Listing datasets...")
    all_datasets = paginated_request(partial(config.api_client.request, RequestMethod.GET, Routes.datasets), "datasets")
    datasets = [Dataset.model_validate(dataset) for dataset in all_datasets]
    logger.debug(f"Listed all datasets, found {len(datasets)} datasets.")
    return datasets


def get_dataset_content(dataset_id: UUID) -> List[Dict]:
    """
    Get the content of a dataset.

    Returns
    -------
    List[Dict]
        The content of the dataset as a list of dictionaries.
        Each dictionary represents a row in the dataset, with an index and a list of values.
        For example (for a dataset with 2 columns):
        [
            {"index": 0, "values": ["value1", "value2"]},
            {"index": 1, "values": ["value3", "value4"]},
            ...
        ]
    """
    config = BaseConfig.get()
    logger.debug("Getting dataset content...")
    all_rows = paginated_request(
        partial(config.api_client.request, RequestMethod.GET, Routes.dataset_content.format(dataset_id=dataset_id)),
        "rows",
    )
    logger.debug("Got dataset content.")
    return all_rows



from collections import ChainMap
from csv import DictReader, DictWriter
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Dict, List, Tuple, Union
from warnings import warn

from galileo_core.constants.dataset_format import DatasetFormat

DatasetType = Union[Path, str, Dict[str, List[str]], List[Dict[str, str]]]


def transpose_dict_to_rows(dataset: Dict[str, List[str]]) -> List[Dict[str, str]]:
    key_rows = [[{key: value} for value in values] for key, values in dataset.items()]
    return [dict(ChainMap(*key_row)) for key_row in zip(*key_rows)]


def parse_dataset(dataset: DatasetType) -> Tuple[Path, DatasetFormat]:
    """
    Set the dataset path and format.

    If the dataset is a dictionary or list, it will be stored as a temporary file and
    the path to the temporary file will be returned. If the dataset is a path or string,
    the path will be returned. If the dataset is not a dictionary, list, path, or
    string, a ValueError will be raised.

    If the dataset is a dictionary or list then it will be stored as a CSV file.

    Parameters
    ----------
    dataset : DatasetType
        Dataset as provided by the user.

    Returns
    -------
    Tuple[Path, DatasetFormat]
        Path to the dataset file and the dataset format.

    Raises
    ------
    ValueError
        If the dataset is not a dictionary, list, path, or string.
    ValueError
        If the dataset file is not a CSV or Feather file.
    FileNotFoundError
        If the dataset file does not exist.
    """
    # Set the dataset path.
    if isinstance(dataset, (dict, list)):
        dataset_rows = transpose_dict_to_rows(dataset) if isinstance(dataset, dict) else dataset
        with NamedTemporaryFile(mode="wt", suffix=f".{DatasetFormat.csv.name}", delete=False) as dataset_file:
            field_names = dataset_rows[0].keys()
            writer = DictWriter(dataset_file, fieldnames=field_names)
            writer.writeheader()
            writer.writerows(dataset_rows)
            dataset_path = Path(dataset_file.name)
    elif isinstance(dataset, Path):
        dataset_path = dataset
    elif isinstance(dataset, str):
        dataset_path = Path(dataset)
    else:
        raise ValueError(f"Invalid dataset type: {type(dataset)}.")
    if not dataset_path.exists():
        raise FileNotFoundError(f"Dataset file {dataset_path} does not exist.")

    # Set the dataset format.
    if dataset_path.suffix.endswith(DatasetFormat.csv):
        dataset_format = DatasetFormat.csv
        # Check if we have embedding columns.
        reader = DictReader(dataset_path.open("rt", errors="ignore"))
        for column in reader.fieldnames or []:
            if column.endswith("_embedding"):
                warn(
                    f"Column {column} appears to be an embedding column, please use "
                    "a feather file as dataset format instead."
                )
    elif dataset_path.suffix.endswith(DatasetFormat.jsonl):
        dataset_format = DatasetFormat.jsonl
    elif dataset_path.suffix.endswith(DatasetFormat.feather):
        dataset_format = DatasetFormat.feather
    else:
        raise ValueError(f"Dataset file {dataset_path} must be a CSV or Feather file.")

    return dataset_path, dataset_format



from pydantic import UUID4

from galileo_core.constants.request_method import RequestMethod
from galileo_core.constants.routes import Routes
from galileo_core.helpers.logger import logger
from galileo_core.schemas.base_config import BaseConfig
from galileo_core.schemas.core.collaboration_role import CollaboratorRole
from galileo_core.schemas.core.user_project import UserProjectCollaboratorRequest, UserProjectCollaboratorResponse


def share_project_with_user(
    project_id: UUID4, user_id: UUID4, role: CollaboratorRole = CollaboratorRole.viewer
) -> UserProjectCollaboratorResponse:
    config = BaseConfig.get()
    logger.debug(f"Sharing project {project_id} with user {user_id} with role {role}...")
    response_dict = config.api_client.request(
        RequestMethod.POST,
        Routes.project_users.format(project_id=project_id),
        json=[UserProjectCollaboratorRequest(user_id=user_id, role=role).model_dump(mode="json")],
    )
    user_shared = [UserProjectCollaboratorResponse.model_validate(user) for user in response_dict]
    logger.debug(f"Shared project {project_id} with user {user_id} with role {role}.")
    return user_shared[0]
